package Mentat::Message::Aggregator;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Message::Aggregator - Short module description

=head1 SYNOPSIS

 use Mentat::Message::Aggregator;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Config::Parser;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant OP_SINGLE => 'SINGLE';
use constant OP_MULTI  => 'MULTI';

use constant MODE_ALTER => 'ALTER';
use constant MODE_BUILD => 'BUILD';
use constant MODE_CLONE => 'CLONE';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Meta structure describing available aggregations
my %AGGREGATIONS = (
        'sum'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l + $r); }],
        'sub'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l - $r); }],
        'mul'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l * $r); }],
        'div'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l / $r); }],
        'mod'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l % $r); }],
        'con'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ("$l, $r"); }],
        'min'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l <= $r)?$l:$r; }],
        'max'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l >= $r)?$l:$r; }],
        'inc'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l + 1); }],
        'join' => [OP_MULTI(),  sub { my ($l, $r) = @_; return [@$l, @$r]; }],
        'frst' => [OP_SINGLE(), sub { my ($l, $r) = @_; return $l; }],
        'last' => [OP_SINGLE(), sub { my ($l, $r) = @_; return $r; }],
    );

# Parameters for rule parser
my $RULE_PARSER_PARAMS;

# Directive processor callback, split the value of the 'mpath' directive key
# to the MPath and value
sub directive_processor
{
    my ($section_name, $section_id, $directive, $value) = @_;
    my ($error, $errmsg) = (0, undef);

    # There should be no other section names but 'ruleset' in the config file
    unless ($section_name eq 'ruleset') { return ($directive, $value, $error, $errmsg); }

    # Split 'rule' directive to aggregation operation and paths
    if ($directive eq 'rule') {
        my ($op,@paths) = split(/:/, $value);
        $value = [$op,@paths];

        unless (exists $AGGREGATIONS{$op}) {
            ($error, $errmsg) = (1, "Invalid aggregation operation '$op' in $section_name, $section_id");
        }
    }
    return ($directive, $value, $error, $errmsg);
}

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;

    # Parameters for rule parser
    $RULE_PARSER_PARAMS = {
        'cb_directive_processor'    => \&directive_processor,
        'stack_values'              => "rule",
        'compress'                  => 0,
        'unique_sections'           => 1,
        'structure'                 => {__general__ => {},
                                        ruleset => {
                                                rule => 'string',
                                            },
                                        },
            };
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Message::Aggregator->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Message::Aggregator reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item load_rulesets($) [PUBLIC,HYBRID]

  Usage     : my $ruleset = $instance->load_rulesets($parser_config)
  Purpose   : Load aggregation rulesets from given external file
  Arguments :
  Returns   :
  Throws    :

=cut

sub load_rulesets($)
{
    my $invocant = shift;
    my ($parser_config,) = @_;

    my $parser  = Config::Parser->new($RULE_PARSER_PARAMS);
    my $config  = $parser->config($parser_config);

    die "Rule file contains errors: " . join(',', @{$parser->get_errors()}) . "\n" unless $config;

    print STDERR Dumper($config);

    return $config->{ruleset};
}

=item aggregate($$) [PUBLIC]

  Usage     : my $result = $instance->aggregate($ruleset_id, @messages);
  Purpose   : Aggregate list of messages into the first one
  Arguments : ARRAY of Mentat::Messages
  Returns   : Mentat::Message reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
              Croaks on aggregation failure

=cut

sub aggregate($$)
{
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
    my ($ruleset_id, @messages) = @_;
    croak ((caller(0))[3] . ": you have to suply at least two messages for aggregation") unless scalar(@messages) >= 2;

    my $target = shift @messages;
    my $result = ($self->{MODE} eq MODE_CLONE())?(clone($target)):($target);
    foreach my $message (@messages) {
        $result = $self->aggregate_into($ruleset_id, $result, $message);
    }
    return $result;
}

=item aggregate_into($$$) [PUBLIC]

  Usage     : my $result = $instance->aggregate_into($target, $source, $ruleset_id);
  Purpose   : Aggregate list of messages into the first one
  Arguments : ARRAY of Mentat::Messages
  Returns   : Mentat::Message reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
              Croaks on aggregation failure

=cut

sub aggregate_into($$$)
{
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
    my ($ruleset_id, $target, $source) = @_;
    croak ((caller(0))[3] . ": you have to suply aggregation ruleset id") unless $ruleset_id;
    croak ((caller(0))[3] . ": you have to suply messages for aggregation") unless $target and $source;

    my $ruleset = $self->_ruleset($ruleset_id);
    my ($type, $left, $right, $aggregation, $src_path, $dst_path, $value);
    foreach my $rule (@{$ruleset->{rule}}) {
        ($aggregation, $src_path, $dst_path) = @$rule;
        $dst_path = ($dst_path)?$dst_path:$src_path;
        $type = $AGGREGATIONS{$aggregation}->[0];
        if ($type eq OP_SINGLE()) {
            $left = $target->path_value($src_path);
            $right = $source->path_value($src_path);
        }
        else {
            $left = $target->path_values($src_path);
            $right = $source->path_values($src_path);
        }
        $value = $self->_aggregate($aggregation, $left, $right);
        ### Setting value: $value
        ### Setting value to path: $dst_path
        $target->path_node_set($dst_path, $value);
    }
    return $target;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Message::Aggregator instance
#  Arguments :
#  Returns   : Mentat::Message::Aggregator reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
    my ($ruleset_config, $immutable) = @_;

    $ruleset_config = $self->load_rulesets($ruleset_config) unless ref $ruleset_config eq 'HASH';

    $self->{RULESETS} = $ruleset_config;
    $self->{MODE} = ($immutable)?MODE_CLONE():MODE_ALTER();

    return $self;
}

# _ruleset($) [PROTECTED]
#
#  Usage     : my $ruleset = $instance->_ruleset($id);
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _ruleset($)
{
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
    my ($id,) = @_;
    croak ((caller(0))[3] . ": aggregation ruleset id must be given as argument") unless $id;
    croak ((caller(0))[3] . ": aggregation ruleset '$id' does not exist") unless exists $self->{RULESETS}->{$id};

    return $self->{RULESETS}->{$id};
}

# _aggregate($) [PROTECTED]
#
#  Usage     : my $aggregate = $instance->_aggregate($ag, $left, $right);
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _aggregate($$$)
{
    my $invocant = shift;
    my ($ag, $left, $right) = @_;
    croak ((caller(0))[3] . ": left and right operands and aggregator must be given as arguments") unless (defined($ag) and defined($left) and defined($right));
    croak ((caller(0))[3] . ": invalid aggregator '$ag'") unless exists $AGGREGATIONS{$ag};

    return &{$AGGREGATIONS{$ag}->[1]}($left, $right);
}

=pod

=back

=cut

1;
