package Mentat::Processor::Engine::Pickup;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor::Engine::Pickup - Mentat message source engine for pickup folder

=head1 SYNOPSIS

 use Mentat::Processor::Engine::Pickup;

=head1 DESCRIPTION

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Time::HiRes qw(time sleep);
use Digest::MD5 qw(md5_hex);
use DirHandle;
use File::Basename;

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Mentat::Core qw(:c_general);
use Mentat::Processor::Engine;
use Mentat::Message::Factory;
use Mentat::Processor::Queue::Folder;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant INTERVAL_STATS => 60;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Processor::Engine');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item eh_process(;$) [PUBLIC]

  Usage     :
  Purpose   : Perform the module processing for given amount of time
  Arguments :
  Returns   :
  Throws    :
  Comments  :
  See Also  :

=cut

sub eh_process(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($last_time,) = @_;
    $last_time = $last_time || time;
    my $alert_message_file;
    my $count = 0;

    # Check, if the DONE queue is not overwhelmed
    unless ($self->_check_queue()) {
        Log::Loger->warn(__PACKAGE__, "[ENG:Pickup] Target queue reached limit '".$self->{QUEUE_CHECK_LIMIT}."', skipping the processing stage and waiting for 15 seconds");
        $self->wait(15.01);

        # Schedule next message handling phase
        $self->emit_event('process', time);
        return (EC_CONTINUE);
    }

    # Check, if the processing is paused
    if ($self->is_paused) {
        Log::Loger->warn(__PACKAGE__, "[ENG:Pickup] Processing paused, skipping the processing stage");

        # Schedule next message handling phase
        $self->emit_event('process', time);
        return (EC_CONTINUE);
    }

    # Scan the pickup folder for new messages
    my ($alert, $queue_id, $request, $result, @args);
    while ($alert_message_file = $self->{QUEUE}->dequeue()) {
        $self->statistics->{'cnt_files_processed'}++;

        Log::Loger->debug(__PACKAGE__, "[ENG:Pickup] Processing message file '$alert_message_file->[1]'");

        # Untaint
        $alert_message_file->[0] = $1 if ($alert_message_file->[0] =~ /(.*)/);
        $alert_message_file->[1] = $1 if ($alert_message_file->[1] =~ /(.*)/);

        $alert = Mentat::Message::Factory->from_file_quiet($alert_message_file->[1]);

        # In case parsing of the file was successfull
        if (defined($alert)) {
            $count++;

            $request = $self->message_create();
            $request->attribute('alert_obj', $alert);
            $request->attribute('queue_id', $alert_message_file->[0]);
            $request->attribute('queue_file', $alert_message_file->[1]);
            Log::Loger->debug(__PACKAGE__, "[ENG:Pickup] Picked up message '$alert_message_file->[0]' from queue");

            $self->emit_event('message_alert', $request);
        }
        # If there was some problem during parsing the file, move the file to the error folder
        else {
            my $result = $self->{QUEUE}->banish($alert_message_file->[0]);

            $self->statistics->{'cnt_errors_parse'}++;
            Log::Loger->error(__PACKAGE__, "[ENG:Pickup] Message '$alert_message_file->[0]' triggered parse error and was banished to $result->[1]");
        }

        last if $count > $self->{PICKUP_LIMIT};
    }
    # Watched directory is empty
    unless ($count) {
        Log::Loger->debug(__PACKAGE__, "[ENG:Pickup] Nothing to do during the processing stage");

        # If we are running in one shot mode, we should terminate when the pickup folder is empty
        # and there are no requests incoming or outgoing
        if ($self->{ONE_SHOT}) {
            my $msg_cnt = $self->{QUEUE}->count_pending(TRUE);
            unless ($msg_cnt) {
                Log::Loger->notice(__PACKAGE__, "[ENG:Pickup] Running in 'ONE SHOT' mode: pickup folder and queues are all empty...terminating");
                $self->done(1);
            }
            else {
                Log::Loger->notice(__PACKAGE__, "[ENG:Pickup] Running in 'ONE SHOT' mode: pickup folder is empty, but still waiting for '$msg_cnt' messages to complete");
            }
        }
        #elsif ((time - $last_time) < 0.01) {
        else {
            # Going too fast? Do not hold CPU
            Log::Loger->notice(__PACKAGE__, "[ENG:Pickup] Incoming queue empty, waiting for 15 seconds");
            $self->wait(15.01);
        }
    }

    # Schedule next message handling phase
    $self->emit_event('process', time);

    Log::Loger->debug(__PACKAGE__, "[ENG:Pickup] Performed '$count' iterations within pickup limit '".$self->{PICKUP_LIMIT}."'");
    return (EC_CONTINUE);
}

=item eh_msg_commit($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :

=cut

sub eh_msg_commit($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message,) = @_;
    confess "Message must be a 'Mentat::Processor::Message' implementation" unless blessed($message) and $message->isa('Mentat::Processor::Message');

    my $queue_id = $message->queue_id;
    my $result = $self->{QUEUE}->commit($queue_id);
    unless ($result) {
        Log::Loger->error(__PACKAGE__, "[ENG:Pickup] Message '$queue_id' was not found in queue");
        return (EC_CONTINUE, $message);
    }

    $self->statistics->{'cnt_files_committed'}++;
    Log::Loger->debug(__PACKAGE__, "[ENG:Pickup] Committed message '$queue_id' [#".$self->statistics->{'cnt_files_committed'}."]");
    return (EC_CONTINUE, $message);
}

=item eh_msg_banish($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :

=cut

sub eh_msg_banish($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message,) = @_;
    confess "Message must be a 'Mentat::Processor::Message' implementation" unless blessed($message) and $message->isa('Mentat::Processor::Message');

    my $queue_id = $message->queue_id;
    my $result = $self->{QUEUE}->banish($queue_id);
    unless ($result) {
        Log::Loger->error(__PACKAGE__, "[ENG:Pickup] Message '$queue_id' was not found in queue");
        return (EC_CONTINUE, $message);
    }

    $self->statistics->{'cnt_files_banished'}++;
    Log::Loger->warn(__PACKAGE__, "[ENG:Pickup] Banished message '$queue_id' from queue [#".$self->statistics->{'cnt_files_banished'}."]");
    return (EC_CONTINUE, $message);
}

=item eh_msg_remove($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :

=cut

sub eh_msg_remove($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message,) = @_;
    confess "Message must be a 'Mentat::Processor::Message' implementation" unless blessed($message) and $message->isa('Mentat::Processor::Message');

    my $queue_id = $message->queue_id;
    my $result = $self->{QUEUE}->remove($queue_id);
    unless ($result) {
        Log::Loger->error(__PACKAGE__, "[ENG:Pickup] Message '$queue_id' was not found in queue");
        return (EC_CONTINUE, $message);
    }

    $self->statistics->{'cnt_files_removed'}++;
    Log::Loger->debug(__PACKAGE__, "[ENG:Pickup] Removed message '$queue_id' from queue [#".$self->statistics->{'cnt_files_removed'}."]");
    return (EC_CONTINUE, $message);
}

=item eh_print_stats($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :

=cut

sub eh_print_stats($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my $speed_processed = (($self->statistics->{'cnt_files_processed'} || 0) - ($self->{STATS_STATE}->{'cnt_files_processed'} || 0)) / INTERVAL_STATS;
    $self->{STATS_STATE}->{'cnt_files_processed'} = ($self->statistics->{'cnt_files_processed'} || 0);
    Log::Loger->warn(__PACKAGE__, "[ENG:Pickup] Processing statistics:\n\tpickup:cnt_files_processed: ".($self->statistics->{'cnt_files_processed'} || 0)."\n".
                                                                       "\tpickup:cnt_files_committed: ".($self->statistics->{'cnt_files_committed'} || 0)."\n".
                                                                       "\tpickup:cnt_files_banished:  ".($self->statistics->{'cnt_files_banished'}  || 0)."\n".
                                                                       "\tpickup:cnt_files_removed:   ".($self->statistics->{'cnt_files_removed'}   || 0)."\n".
                                                                       "\tpickup:spd_files_processed: ".sprintf('%.2f', $speed_processed)." #/s"
    );
    $self->emit_event_at('print_stats', (time + INTERVAL_STATS));
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;

    # Call parent version first
    $self = $self->SUPER::_init(%config);

    die "[ENG:Pickup] Constructor expects configurations as argument\n" unless %config;
    die "[ENG:Pickup] Pickup engine configuration 'watched_folder' is missing\n" unless $config{watched_folder};
    die "[ENG:Pickup] Pickup engine configuration 'queue_folder' is missing\n" unless $config{queue_folder};
    die "[ENG:Pickup] Pickup engine configuration 'temp_folder' is missing\n" unless $config{temp_folder};
    die "[ENG:Pickup] Pickup engine configuration 'error_folder' is missing\n" unless $config{error_folder};
    die "[ENG:Pickup] Pickup engine configuration 'pickup_limit' is missing\n" unless $config{pickup_limit};

    $self->{QUEUE} = Mentat::Processor::Queue::Folder->new(
                        incoming_folder => $config{watched_folder},
                        pending_folder  => $config{queue_folder},
                        done_folder     => $config{done_folder},
                        temp_folder     => $config{temp_folder},
                        error_folder    => $config{error_folder});
    $self->{PICKUP_LIMIT} = $config{pickup_limit};
    $self->{ONE_SHOT}     = $config{one_shot};

    $self->{QUEUE_CHECK_TIME}  = undef;
    $self->{QUEUE_CHECK_LIMIT} = $config{queue_check_limit} || 10000;

    $self->{STATS_STATE} = {};

    Log::Loger->info(__PACKAGE__, "[ENG:Pickup] Watching folder '$config{watched_folder}' with pickup limit '$config{pickup_limit}'");
    Log::Loger->info(__PACKAGE__, "[ENG:Pickup] Running in 'ONE SHOT' mode, will terminate when pickup folder is empty") if $self->{ONE_SHOT};
    Log::Loger->info(__PACKAGE__, "[ENG:Pickup] Using folder '$config{queue_folder}' as PENDING folder");
    if ($config{done_folder}) {
        Log::Loger->info(__PACKAGE__, "[ENG:Pickup] Using folder '$config{done_folder}' as DONE folder with limit '$self->{QUEUE_CHECK_LIMIT}'");
    }
    else {
        Log::Loger->info(__PACKAGE__, "[ENG:Pickup] No DONE folder specified, messages will be permanently deleted after processing");
    }
    Log::Loger->info(__PACKAGE__, "[ENG:Pickup] Using folder '$config{temp_folder}' as TEMP folder");
    Log::Loger->info(__PACKAGE__, "[ENG:Pickup] Using folder '$config{error_folder}' as ERROR folder");

    # Check, if the queue is empty
    if ($self->{QUEUE}->count_pending()) {
        Log::Loger->warn(__PACKAGE__, "[ENG:Pickup] Pickup queue '".$config{queue_folder}."' is not empty, move the messages manually to pickup folder '".$config{watched_folder}."'");
    }

    return $self;
}

# _get_events_emitted(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of events emited by this component
#  Arguments : LIST OF ARRAY REFERENCES @lists - event lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $events containing event names as STRINGS
#  Throws    : Dies, if not invoked on instance

sub _get_events_emitted(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;

    return $self->SUPER::_get_events_emitted([
        'message_alert', 'message_commit', 'message_banish', 'message_remove', 'process', 'print_stats'
    ], @lists);
}

# _get_events_handled(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of events and appropriate handlers for this component
#  Arguments : LIST OF ARRAY REFERENCES @lists - event handler lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $handlers with internal structure [$event, $handler]
#                  STRING $event   - event name
#                  STRING $handler - event handler name
#  Throws    : Dies, if not invoked on instance

sub _get_events_handled(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;

    return $self->SUPER::_get_events_handled([
        ['message_commit', 'eh_msg_commit'],
        ['message_banish', 'eh_msg_banish'],
        ['message_remove', 'eh_msg_remove'],
        ['print_stats',    'eh_print_stats'],
    ], @lists);
}

sub _check_queue()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $ct = time;

    # Since we are retrieving messages one by one, perform the check only every 30 seconds
    if (not $self->{QUEUE_CHECK_TIME} or ($self->{QUEUE_CHECK_TIME} + 30) < $ct) {
        $self->{QUEUE_CHECK_TIME} = $ct;
        #Log::Loger->info(__PACKAGE__, "[ENG:Pickup] Checking DONE folder size");
        if ($self->{QUEUE}->count_done() >= $self->{QUEUE_CHECK_LIMIT}) {
            return 0;
        }
    }
    return 1;
}

=pod

=back

=cut

1;
